# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This code is based on the code for the AWS CLI v2's `aws sso login` functionality
# https://github.com/aws/aws-cli/tree/v2/awscli/customizations/sso

import argparse
import os
import sys
import json
import logging
import datetime

from botocore.session import Session
from botocore.exceptions import ClientError

import click

from ..sso import get_token_loader, get_credentials
from ..exceptions import InvalidSSOConfigError, AuthDispatchError, InteractiveAuthDisabledError

LOG_FILE = os.path.expanduser(
    os.path.join('~', '.aws', 'sso', 'aws-sso-credential-process-log.txt')
)

LOGGER = logging.getLogger(__name__)


CONFIG_VARS = [
    ('start url', 'sso_start_url'),
    ('SSO region', 'sso_region'),
    ('role', 'sso_role_name'),
    ('account', 'sso_account_id'),
    ('interactive', 'sso_interactive_auth')
]

def get_config(arg_config, profile_config, logger=None):
    sso_config = {}
    missing_vars = []
    for friendly_name, config_var_name in CONFIG_VARS:
        if arg_config.get(config_var_name):
            sso_config[config_var_name] = arg_config[config_var_name]
        elif config_var_name not in profile_config:
            missing_vars.append((friendly_name, config_var_name))
            sso_config[config_var_name] = None
        else:
            sso_config[config_var_name] = profile_config[config_var_name]

    required_vars = ['sso_start_url', 'sso_region', 'sso_account_id', 'sso_role_name']

    # TODO: enable when interactive account and role picker is implemented
    # interactive = sso_config.get('sso_interactive_auth') == 'true'
    # if interactive:
    #     required_vars = ['sso_start_url', 'sso_region']

    missing_requred_vars = [v[0] for v in missing_vars if v[1] in required_vars]
    if missing_requred_vars:
        raise InvalidSSOConfigError(
            'Missing ' + ', '.join(missing_requred_vars)
        )
    return sso_config


@click.command('credential-process')
@click.option('--profile', help='Extract settings from the given profile')
@click.option('--role-name')
@click.option('--account-id')
@click.option('--start-url')
@click.option('--region')

@click.option('--force-refresh', is_flag=True, help='Do not reuse cached AWS SSO token')
@click.option('--debug', is_flag=True, help='Write to the debugging log file')
def credential_process(
        profile,
        role_name,
        account_id,
        start_url,
        region,
        force_refresh,
        debug):

    if debug or os.environ.get('AWS_SSO_CREDENTIAL_PROCESS_DEBUG', '').lower() in ['1', 'true']:
        logging.basicConfig(level=logging.DEBUG, filename=LOG_FILE, filemode='w')
    else:
        logging.disable(logging.CRITICAL)

    LOGGER.info('Starting credential process at {}'.format(datetime.datetime.now().isoformat()))

    if role_name is None and os.environ.get('AWS_SSO_ROLE_NAME'):
        LOGGER.debug("Using role from env: {}".format(os.environ['AWS_SSO_ROLE_NAME']))
        role_name = os.environ['AWS_SSO_ROLE_NAME']

    if account_id is None and os.environ.get('AWS_SSO_ACCOUNT_ID'):
        LOGGER.debug("Using acccount from env: {}".format(os.environ['AWS_SSO_ACCOUNT_ID']))
        account_id = os.environ['AWS_SSO_ACCOUNT_ID']

    # if role_name and role_name.startswith('arn'):
    #     parts = role_name.split(':')
    #     account_id = parts[4]
    #     role_name = parts[5].split('/', 1)[1]

    if start_url is None and os.environ.get('AWS_SSO_START_URL'):
        start_url = os.environ['AWS_SSO_START_URL']

    if region is None and os.environ.get('AWS_SSO_REGION'):
        region = os.environ['AWS_SSO_REGION']

    session_kwargs = {}

    if profile:
        session_kwargs['profile'] = profile

    arg_config = {
        'sso_start_url': start_url,
        'sso_region': region,
        'sso_role_name': role_name,
        'sso_account_id': account_id,
    }

    LOGGER.info('CONFIG FROM ARGS: {}'.format(json.dumps(arg_config)))

    try:
        session = Session(**session_kwargs)

        if profile:
            profile_config = session.get_scoped_config()
            LOGGER.info('CONFIG FROM PROFILE: {}'.format(json.dumps(profile_config)))
        else:
            profile_config = {}

        config = get_config(arg_config, profile_config, logger=LOGGER)

        LOGGER.info('CONFIG: {}'.format(json.dumps(config)))

        if (config.get('sso_interactive_auth') or '').lower() == 'true':
            raise InvalidSSOConfigError('Interactive auth has been removed. See https://github.com/benkehoe/aws-sso-credential-process/issues/4')

        token_loader = get_token_loader(
            session=session,
            sso_region=config['sso_region'],
            force_refresh=force_refresh,
            logger=LOGGER,
        )

        if not config['sso_account_id']:
            #TODO: if interactive, prompt for account
            raise InvalidSSOConfigError('Missing account id')

        if not config['sso_role_name']:
            #TODO: if interactive, prompt for role
            raise InvalidSSOConfigError('Missing role')

        credentials = get_credentials(
            session=session,
            sso_region=config['sso_region'],
            start_url=config['sso_start_url'],
            account_id=config['sso_account_id'],
            role_name=config['sso_role_name'],
            token_loader=token_loader,
            logger=LOGGER,
        )

        output = {
            "Version": 1,
            "AccessKeyId": credentials['access_key'],
            "SecretAccessKey": credentials['secret_key'],
            "SessionToken": credentials['token'],
            # as provided the expiration isn't valid ISO8601 and that causes parsing errors for some SDKs
            "Expiration": credentials['expiry_time'].replace('UTC', 'Z'),
        }
        LOGGER.debug('CREDENTIALS: ' + json.dumps(output))

        print(json.dumps(output, separators=(',', ':')))
    except InteractiveAuthDisabledError as e:
        LOGGER.info('Auth needed but interactive auth disabled')
        print('Interactive auth disabled, use `aws sso login` and try again', file=sys.stderr)
        sys.exit(1)
    except InvalidSSOConfigError as e:
        LOGGER.error(e)
        print(e, file=sys.stderr)
        sys.exit(2)
    except AuthDispatchError as e:
        LOGGER.error(e)
        print(e, file=sys.stderr)
        sys.exit(3)
    except ClientError as e:
        LOGGER.error(e, exc_info=True)
        #TODO: print a different message for AccessDeniedException during CreateToken? -> user canceled login
        # boto_error_matches(e, 'CreateToken', 'AccessDeniedException')
        print('ERROR:', e, file=sys.stderr)
        sys.exit(4)
    except Exception as e:
        LOGGER.error(e, exc_info=True)
        print('ERROR:', e, file=sys.stderr)
        sys.exit(5)

if __name__ == '__main__':
    credential_process(prog_name="python -m aws_sso_util.cli.credential_process")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



import logging

import boto3

LOGGER = logging.getLogger(__name__)

class RetrievalError(Exception):
    pass

PROFILE = None
SESSION = None
def get_session(logger=None):
    if not logger:
        logger = LOGGER
    else:
        logger = logger.getChild("api_utils")

    global PROFILE, SESSION
    if not SESSION:
        if PROFILE:
            logger.debug(f"Creating session with profile {PROFILE}")
        else:
            logger.debug(f"Creating session")
        SESSION = boto3.Session(profile_name=PROFILE)
    return SESSION

SSO_INSTANCE = None
def get_sso_instance(session=None, logger=None):
    if not session:
        session = get_session(logger=logger)

    if not logger:
        logger = LOGGER
    else:
        logger = logger.getChild("api_utils")

    global SSO_INSTANCE
    if not SSO_INSTANCE:
        logger.info(f"Retrieving SSO instance")
        response = get_session().client('sso-admin').list_instances()
        if len(response['Instances']) == 0:
            raise RetrievalError("No SSO instance found, must specify instance")
        elif len(response['Instances']) > 1:
            raise RetrievalError("{} SSO instances found, must specify instance".format(len(response['Instances'])))
        else:
            instance_arn = response['Instances'][0]['InstanceArn']
            instance_id = instance_arn.split('/')[-1]
            logger.info("Using SSO instance {}".format(instance_id))
            SSO_INSTANCE = instance_arn
    if SSO_INSTANCE.startswith('ssoins-') or SSO_INSTANCE.startswith('ins-'):
        SSO_INSTANCE = f"arn:aws:sso:::instance/{SSO_INSTANCE}"
    return SSO_INSTANCE

OU_CACHE = {}
def get_accounts_for_ou(ou, recursive, refresh=False, session=None, cache=None, logger=None):
    if not session:
        session = get_session(logger=logger)

    if not logger:
        logger = LOGGER
    else:
        logger = logger.getChild("api_utils")

    global OU_CACHE
    if not cache:
        cache = OU_CACHE

    ou_children_key = f"{ou}#children"
    ou_accounts_key = f"{ou}#accounts"

    accounts = []

    if refresh or ou_accounts_key not in cache:
        logger.info(f"Retrieving accounts for OU {ou}")
        client = session.client('organizations')

        ou_accounts = []

        paginator = client.get_paginator('list_accounts_for_parent')
        for response in paginator.paginate(ParentId=ou):
            ou_accounts.extend(data['Id'] for data in response['Accounts'])

        cache[ou_accounts_key] = ou_accounts

        accounts.extend(ou_accounts)
    else:
        accounts.extend(cache[ou_accounts_key])

    if recursive:
        if refresh or ou_children_key not in cache:
            logger.info(f"Retrieving child OUs for OU {ou}")
            client = session.client('organizations')

            ou_children = []

            paginator = client.get_paginator('list_organizational_units_for_parent')
            for response in paginator.paginate(ParentId=ou):
                sub_ous = [data['Id'] for data in response['OrganizationalUnits']]
                ou_children.extend(sub_ous)

            cache[ou_children_key] = ou_children

        for sub_ou_id in cache[ou_children_key]:
            accounts.extend(get_accounts_for_ou(sub_ou_id, True, refresh=refresh, session=session, cache=cache, logger=logger))

    return accounts

class LookupError(Exception):
    pass

class Ids:
    def __init__(self, session, instance_arn, identity_store_id):
        self._client = session.client('sso-admin')
        self._instance_arn = instance_arn
        self._instance_arn_printed = False
        self._identity_store_id = identity_store_id
        self._identity_store_id_printed = False
        self.suppress_print = False

    def _print(self, *args, **kwargs):
        if not self.suppress_print:
            print(*args, **kwargs)

    @property
    def instance_arn(self):
        if self._instance_arn:
            if not self._instance_arn_printed:
                self._print("Using SSO instance {}".format(self._instance_arn.split('/')[-1]))
                self._instance_arn_printed = True
            return self._instance_arn
        response = self._client.list_instances()
        if len(response['Instances']) == 0:
            raise LookupError("No SSO instance found, please specify with --instance-arn")
        elif len(response['Instances']) > 1:
            raise LookupError("{} SSO instances found, please specify with --instance-arn".format(len(response['Instances'])))
        else:
            instance_arn = response['Instances'][0]['InstanceArn']
            self._instance_arn = instance_arn
            self._print("Using SSO instance {}".format(self._instance_arn.split('/')[-1]))
            self._instance_arn_printed = True
            identity_store_id = response['Instances'][0]['IdentityStoreId']
            if self._identity_store_id and self._identity_store_id != identity_store_id:
                raise LookupError("SSO instance identity store {} does not match given identity store {}".format(identity_store_id, self._identity_store_id))
            else:
                self._identity_store_id = identity_store_id
        return self._instance_arn

    @property
    def identity_store_id(self):
        if self._identity_store_id:
            if not self._identity_store_id_printed:
                self._print("Using SSO identity store {}".format(self._identity_store_id))
                self._identity_store_id_printed = True
            return self._identity_store_id
        response = self._client.list_instances()
        if len(response['Instances']) == 0:
            raise LookupError("No SSO instance found, please specify identity store with --identity-store-id or instance with --instance-arn")
        elif len(response['Instances']) > 1:
            raise LookupError("{} SSO instances found, please specify identity store with --identity-store-id or instance with --instance-arn".format(len(response['Instances'])))
        else:
            identity_store_id = response['Instances'][0]['IdentityStoreId']
            self._identity_store_id = identity_store_id
            self._print("Using SSO identity store {}".format(self._identity_store_id))
            self._identity_store_id_printed = True
            instance_arn = response['Instances'][0]['InstanceArn']
            instance_id = instance_arn.split('/')[-1]
            if self._instance_arn and self._instance_arn != instance_arn:
                raise LookupError("SSO instance {} does not match given instance {}".format(instance_id, self._instance_arn.split('/')[-1]))
            else:
                self._instance_arn = instance_arn
        return self._identity_store_id



import click

from .. import __version__

from .cfn import generate_template
from .configure_profile import configure_profile
from .credential_process import credential_process
from .populate_profiles import populate_profiles
from .lookup import lookup
from .assignments import assignments

@click.group(name="aws-sso-util")
@click.version_option(version=__version__, message='%(version)s')
def cli():
    pass

@cli.group()
def configure():
    pass

configure.add_command(configure_profile, "profile")
configure.add_command(populate_profiles, "populate")

cli.add_command(lookup)
cli.add_command(assignments)

cli.add_command(generate_template, "cfn")

cli.add_command(credential_process)




import argparse
from collections import namedtuple, OrderedDict
from pathlib import Path
import logging
import sys
import os

import boto3

import click

from .. import api_utils
from ..cfn.config import Config, ConfigError, validate_config
from ..cfn import resources, templates, macro, utils

Input = namedtuple("Input", ["base_path", "stem", "assignments", "permission_sets", "base_template"])

TemplateProcessInput = namedtuple("TemplateProcessInput", [
    "base_path",
    "base_stem",
    "base_template",
    "max_stack_resources",
    "items"
])
TemplateProcessInputItem = namedtuple("TemplateProcessInput", [
    "stem",
    "resource_collection",
])

LOGGER = logging.getLogger("sso_cfn")

def param_loader(ctx, param, value):
    if not value:
        return {}
    d = []
    for p in value.split(","):
        if "=" in p:
            d.append((p.split("=", 1)))
        else:
            d.append((p, None))
    return dict(d)

@click.command()
@click.argument("config_file", required=True, nargs=-1, type=click.File("r"))

@click.option("--macro", is_flag=True, help="Process templates with macro")

@click.option("--profile", help="AWS profile to use to retrieve SSO instance and/or accounts from OUs")
@click.option("--sso-instance", "--ins", help="If not provided, will be retrieved from your account")

@click.option("--template-file-suffix")
@click.option("--output-dir")

@click.option("--base-template-file", type=click.File("r"), help="Base template to build from")
@click.option("--template-parameters", callback=param_loader, help="String-type parameters on the template")

@click.option("--max-resources-per-template", type=int, default=templates.MAX_RESOURCES_PER_TEMPLATE)
@click.option("--max-concurrent-assignments", type=int, default=templates.MAX_CONCURRENT_ASSIGNMENTS)

@click.option("--verbose", "-v", count=True)

def generate_template(
        config_file,
        macro,
        profile,
        sso_instance,
        template_file_suffix,
        output_dir,
        base_template_file,
        template_parameters,
        max_resources_per_template,
        max_concurrent_assignments,
        verbose):
    logging.basicConfig(format="{message}", style="{")
    if verbose >= 1:
        LOGGER.setLevel(logging.DEBUG)
    else:
        LOGGER.setLevel(logging.INFO)
    if verbose == 2:
        logging.getLogger().setLevel(logging.INFO)
    elif verbose == 3:
        logging.getLogger().setLevel(logging.DEBUG)

    if macro and base_template_file:
        raise click.UsageError("--base-template-file not allowed with --macro")
    if macro and template_parameters:
        raise click.UsageError("--template-parameters not allowed with --macro")

    api_utils.PROFILE = profile
    api_utils.SSO_INSTANCE = sso_instance

    if not template_file_suffix:
        template_file_suffix = ".yaml"
    elif not template_file_suffix.endswith(".yaml"):
        template_file_suffix = template_file_suffix + ".yaml"

    if base_template_file:
        base_template = utils.load_yaml(base_template_file)
        base_template_path = Path(base_template_file.name).resolve()
        prev_len = len(config_file)
        config_file = [c for c in config_file if Path(c.name).resolve() != base_template_path]
        if len(config_file) != prev_len:
            LOGGER.debug("Removed base template file from list of config files")
    else:
        base_template = None

    if macro:
        template_process_inputs = process_macro(
            config_file,
            sso_instance,
            template_file_suffix,
            output_dir,
            max_resources_per_template,
            max_concurrent_assignments
        )
    else:
        template_process_inputs = process_config(
            config_file,
            sso_instance,
            template_file_suffix,
            output_dir,
            base_template,
            max_resources_per_template,
            max_concurrent_assignments
        )

    templates_to_write = process_templates(
        template_process_inputs,
        template_file_suffix,
        max_resources_per_template,
        max_concurrent_assignments)

    write_templates(templates_to_write)

def process_config(
    config_file,
    sso_instance,
    template_file_suffix,
    output_dir,
    base_template,
    max_resources_per_template,
    max_concurrent_assignments):
    template_process_inputs = {}

    for config_file_fp in config_file:
        LOGGER.info(f"Loading config file {config_file_fp.name}")
        config_file_path = Path(config_file_fp.name)
        if output_dir:
            base_path = Path(output_dir)
        else:
            base_path = config_file_path.parent / "templates"
        stem = config_file_path.stem

        data = utils.load_yaml(config_file_fp)
        LOGGER.debug(f"Config file contents:\n{utils.dump_yaml(data)}")

        config = Config()
        config.load(data, logger=LOGGER)

        try:
            validate_config(config, sso_instance, api_utils.get_sso_instance, logger=LOGGER)
        except ConfigError as e:
            LOGGER.fatal(f"{e!s} in {config_file_path}")
            sys.exit(1)

        resource_collection = resources.get_resources_from_config(
            config,
            ou_fetcher=lambda ou, recursive: api_utils.get_accounts_for_ou(ou, recursive, logger=LOGGER),
            logger=LOGGER)

        max_stack_resources = templates.get_max_number_of_child_stacks(
            resource_collection.num_resources,
            max_resources_per_template=max_resources_per_template)

        template_process_inputs[config_file_path] = TemplateProcessInput(
            base_path=base_path,
            base_stem=stem,
            base_template=base_template,
            max_stack_resources=max_stack_resources,
            items=[TemplateProcessInputItem(
                stem=stem,
                resource_collection=resource_collection
            )]
        )
    return template_process_inputs

def process_macro(
        config_file,
        sso_instance,
        template_file_suffix,
        output_dir,
        max_resources_per_template,
        max_concurrent_assignments):

    template_process_inputs = {}

    for config_file_fp in config_file:
        LOGGER.info(f"Loading template file {config_file_fp.name}")
        config_file_path = Path(config_file_fp.name)
        if output_dir:
            base_path = Path(output_dir)
        else:
            base_path = config_file_path.parent / "templates"
        stem = config_file_path.stem

        input_template = utils.load_yaml(config_file_fp)
        LOGGER.debug(f"Input template:\n{utils.dump_yaml(input_template)}")

        LOGGER.info("Extracting resources from template")
        base_template, max_stack_resources, resource_collection_dict = macro.process_template(input_template,
                instance_fetcher=api_utils.get_sso_instance,
                max_resources_per_template=max_resources_per_template,
                logger=LOGGER)

        template_process_inputs[config_file_path] = TemplateProcessInput(
            base_path=base_path,
            base_stem=stem,
            base_template=base_template,
            max_stack_resources=max_stack_resources,
            items=[TemplateProcessInputItem(
                stem=resource_name,
                resource_collection=resource_collection
            ) for resource_name, resource_collection in resource_collection_dict.items()]
        )

    return template_process_inputs

def process_templates(
        template_process_inputs,
        template_file_suffix,
        max_resources_per_template,
        max_concurrent_assignments):
    templates_to_write = {}

    for name, template_process_input in template_process_inputs.items():
        LOGGER.info(f"Generating templates for {name}")
        parent_template_to_write = template_process_input.base_template or {}
        all_children = []
        for template_process_input_item in template_process_input.items:
            num_parent_resources = len(parent_template_to_write.get("Resources", {})) + template_process_input.max_stack_resources

            parent_template = templates.resolve_templates(
                template_process_input_item.resource_collection.assignments,
                template_process_input_item.resource_collection.permission_sets,
                max_resources_per_template=max_resources_per_template,
                num_parent_resources=num_parent_resources,
            )

            template_collection = parent_template.get_templates(
                template_process_input.base_path,
                ".",
                template_process_input_item.stem,
                template_file_suffix,
                base_template=parent_template_to_write,
                parameters=None,
                max_concurrent_assignments=max_concurrent_assignments,
                path_joiner=os.path.join

            )
            parent_template_to_write = template_collection.parent.template
            LOGGER.debug(f"Intermediate parent template\n{utils.dump_yaml(parent_template_to_write)}")

            all_children.extend(template_collection.children)

        templates_to_write[name] = templates.TemplateCollection(
            parent=templates.WritableTemplate(
                path=os.path.join(
                    template_process_input.base_path,
                    f"{template_process_input.base_stem}{template_file_suffix}"),
                template=parent_template_to_write,
            ),
            children=all_children
        )
    return templates_to_write

def write_templates(templates_to_write):
    for name, template_collection_to_write in templates_to_write.items():
        parent_path = template_collection_to_write.parent.path
        parent_data = template_collection_to_write.parent.template

        for child_path, child_data in template_collection_to_write.children:
            LOGGER.info(f"Writing child template at path {child_path}")
            Path(child_path).parent.mkdir(parents=True, exist_ok=True)
            with open(child_path, "w") as fp:
                utils.dump_yaml(child_data, fp)

        LOGGER.info(f"Writing template for {name} at path {parent_path}")
        Path(parent_path).parent.mkdir(parents=True, exist_ok=True)
        with open(parent_path, "w") as fp:
            utils.dump_yaml(parent_data, fp)


if __name__ == "__main__":
    generate_template(prog_name="python -m aws_sso_util.cli.cfn")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



import os
import argparse
import logging
import json
import subprocess
import shlex
import re
from collections import namedtuple

import botocore
from botocore.session import Session
from botocore.exceptions import ClientError, ProfileNotFound

import click

from ..sso import get_token_loader
from ..vendored_botocore.config_file_writer import ConfigFileWriter, write_values, get_config_filename

LOGGER = logging.getLogger(__name__)

ConfigParams = namedtuple("ConfigParams", ["profile_name", "account_name", "account_id", "role_name", "region"])

def get_short_region(region):
    area, direction, num = region.split("-")
    dir_abbr = {
        "north": "no",
        "northeast": "ne",
        "east": "ea",
        "southeast": "se",
        "south": "so",
        "southwest": "sw",
        "west": "we",
        "northwest": "nw",
        "central": "ce",
    }
    return "".join([area, dir_abbr.get(direction, direction), num])

KNOWN_COMPONENTS = [
    "account_name",
    "account_id",
    "account_number",
    "role_name",
    "region",
    "short_region",
]
def generate_profile_name_format(input, separator, region_style):
    def process_component(c):
        if c == "default_style_region":
            if region_style == "short":
                c = "short_region"
            else:
                c = "region"
        if c in KNOWN_COMPONENTS:
            return "{" + c + "}"
        else:
            return c
    region_format = separator.join(process_component(c) for c in input.split(","))
    no_region_format = separator.join(process_component(c) for c in input.split(",") if c not in ["default_style_region", "region", "short_region"])
    return region_format, no_region_format

def get_formatter(include_region, region_format, no_region_format):
    def proc_kwargs(kwargs):
        kwargs["short_region"] = get_short_region(kwargs["region"])
        kwargs["account_number"] = kwargs["account_id"]
        return kwargs
    if include_region == "default":
        def formatter(i, **kwargs):
            kwargs = proc_kwargs(kwargs)
            if i == 0:
                return no_region_format.format(**kwargs)
            else:
                return region_format.format(**kwargs)
        return formatter
    elif include_region == "always":
        def formatter(i, **kwargs):
            kwargs = proc_kwargs(kwargs)
            return region_format.format(**kwargs)
        return formatter
    else:
        raise ValueError("Unknown include_region value {}".format(include_region))

PROCESS_FORMATTER_ARGS = [
    "account_name",
    "account_id",
    "role_name",
    "region",
    "short_region",
    "region_index",
]
def get_process_formatter(command):
    def formatter(i, **kwargs):
        kwargs["region_index"] = str(i)
        kwargs["short_region"] = get_short_region(kwargs["region"])
        run_args = [command]
        for component in PROCESS_FORMATTER_ARGS:
            run_args.append(shlex.quote(kwargs[component]))
        try:
            result = subprocess.run(' '.join(run_args), shell=True, capture_output=True, check=True)
        except subprocess.CalledProcessError as e:
            lines = [
                "Profile name process failed ({})".format(e.returncode)
            ]
            if e.stdout:
                lines.append(e.stdout.decode('utf-8'))
            if e.stderr:
                lines.append(e.stderr.decode('utf-8'))
            LOGGER.error("\n".join(lines))
            raise e
        return result.stdout.decode('utf-8').strip()
    return formatter

def get_trim_formatter(account_name_patterns, role_name_patterns, formatter):
    def trim_formatter(i, **kwargs):
        for pattern in account_name_patterns:
            kwargs["account_name"] = re.sub(pattern, '', kwargs["account_name"])
        for pattern in role_name_patterns:
            kwargs["role_name"] = re.sub(pattern, '', kwargs["role_name"])
        return formatter(i, **kwargs)
    return trim_formatter

@click.command()
@click.option("--sso-start-url")
@click.option("--sso-region")

@click.option("--region", "-r", "regions", multiple=True)

@click.option("--dry-run", is_flag=True, help="Print the config to stdout")

@click.option("--config-default", "-d", multiple=True)
@click.option("--existing-config-action", type=click.Choice(["keep", "overwrite", "discard"]), default="keep")

@click.option("--components", "profile_name_components", default="account_name,role_name,default_style_region")
@click.option("--separator", "--sep", "profile_name_separator", default="_")
@click.option("--include-region", "profile_name_include_region", type=click.Choice(["default", "always"]), default="default")
@click.option("--region-style", "profile_name_region_style", type=click.Choice(["short", "long"]), default="short")
@click.option("--trim-account-name", "profile_name_trim_account_name_patterns", multiple=True, default=[], help="Regex to remove from account names")
@click.option("--trim-role-name", "profile_name_trim_role_name_patterns", multiple=True, default=[], help="Regex to remove from account names")
@click.option("--profile-name-process")

@click.option("--credential-process/--no-credential-process")

@click.option("--force-refresh", is_flag=True, help="Re-login")
@click.option("--debug", is_flag=True)
def populate_profiles(
        sso_start_url,
        sso_region,
        regions,
        dry_run,
        config_default,
        existing_config_action,
        profile_name_components,
        profile_name_separator,
        profile_name_include_region,
        profile_name_region_style,
        profile_name_trim_account_name_patterns,
        profile_name_trim_role_name_patterns,
        profile_name_process,
        credential_process,
        force_refresh,
        debug):

    logging.basicConfig()
    LOGGER.setLevel(logging.DEBUG if debug else logging.INFO)

    missing = []

    if not sso_start_url:
        sso_start_url = os.environ.get("AWS_CONFIGURE_SSO_DEFAULT_SSO_START_URL", os.environ.get("AWS_CONFIGURE_DEFAULT_SSO_START_URL"))
    if not sso_start_url:
        missing.append("--start-url")

    if not sso_region:
        sso_region = os.environ.get("AWS_CONFIGURE_SSO_DEFAULT_SSO_REGION", os.environ.get("AWS_CONFIGURE_DEFAULT_SSO_REGION"))
    if not sso_region:
        missing.append("--sso-region")

    if not regions and os.environ.get("AWS_CONFIGURE_DEFAULT_REGION"):
        regions = [os.environ["AWS_CONFIGURE_DEFAULT_REGION"]]
    if not regions:
        missing.append("--region")

    if missing:
        raise click.UsageError("Missing arguments: {}".format(", ".join(missing)))

    if config_default:
        config_default = dict(v.split("=", 1) for v in config_default)
    else:
        config_default = {}

    if profile_name_process:
        profile_name_formatter = get_process_formatter(profile_name_process)
    else:
        region_format, no_region_format = generate_profile_name_format(profile_name_components, profile_name_separator, profile_name_region_style)
        LOGGER.debug("Profile name format (region):    {}".format(region_format))
        LOGGER.debug("Profile name format (no region): {}".format(no_region_format))
        profile_name_formatter = get_formatter(profile_name_include_region, region_format, no_region_format)
        if profile_name_trim_account_name_patterns or profile_name_trim_role_name_patterns:
            profile_name_formatter = get_trim_formatter(profile_name_trim_account_name_patterns, profile_name_trim_role_name_patterns, profile_name_formatter)

    try:
        profile_name_formatter(0, account_name='foo', account_id='bar', role_name='baz', region='us-east-1')
    except Exception as e:
        raise click.UsageError("Invalid profile name format: {}".format(e))

    session = Session()

    token_loader = get_token_loader(session,
            sso_region,
            interactive=True,
            force_refresh=force_refresh,
            logger=LOGGER)

    LOGGER.info("Logging in")
    token = token_loader(sso_start_url)

    LOGGER.debug("Token: {}".format(token))

    config = botocore.config.Config(
        region_name=sso_region,
        signature_version=botocore.UNSIGNED,
    )
    client = session.create_client("sso", config=config)

    LOGGER.info("Gathering accounts and roles")
    accounts = []
    list_accounts_args = {
        "accessToken": token
    }
    while True:
        response = client.list_accounts(**list_accounts_args)

        accounts.extend(response["accountList"])

        next_token = response.get("nextToken")
        if not next_token:
            break
        else:
            list_accounts_args["nextToken"] = response["nextToken"]

    LOGGER.debug("Account list: {} {}".format(len(accounts), accounts))

    configs = []
    for account in accounts:
        LOGGER.debug("Getting roles for {}".format(account["accountId"]))
        list_role_args = {
            "accessToken": token,
            "accountId": account["accountId"],
        }

        while True:
            response = client.list_account_roles(**list_role_args)

            for role in response["roleList"]:
                for i, region in enumerate(regions):
                    profile_name = profile_name_formatter(i,
                        account_name=account["accountName"],
                        account_id=account["accountId"],
                        role_name=role["roleName"],
                        region=region,
                    )
                    configs.append(ConfigParams(profile_name, account["accountName"], account["accountId"], role["roleName"], region))

            next_token = response.get("nextToken")
            if not next_token:
                break
            else:
                list_role_args["nextToken"] = response["nextToken"]


    configs.sort(key=lambda v: v.profile_name)

    LOGGER.debug("Got configs: {}".format(configs))

    if not dry_run:
        LOGGER.info("Writing {} profiles to {}".format(len(configs), get_config_filename(session)))

        config_writer = ConfigFileWriter()
        def write_config(profile_name, config_values):
            write_values(session, config.profile_name, config_values, config_file_writer=config_writer)
    else:
        LOGGER.info("Dry run for {} profiles".format(len(configs)))
        def write_config(profile_name, config_values):
            lines = [
                "[profile {}]".format(profile_name)
            ]
            for key, value in config_values.items():
                lines.append("{} = {}".format(key, value))
            lines.append("")
            print("\n".join(lines))

    for config in configs:
        LOGGER.debug("Processing config: {}".format(config))

        existing_values = {}
        if existing_config_action != "discard":
            try:
                existing_values = Session(profile=config.profile_name).get_scoped_config()
            except ProfileNotFound:
                pass

        config_values = {
            "sso_account_name": config.account_name,
            "sso_account_id": config.account_id,
            "sso_role_name": config.role_name,
            "sso_start_url": sso_start_url,
            "sso_region": sso_region,
            "region": config.region,
        }


        for k, v in config_default.items():
            if k in existing_values and existing_config_action in ["keep"]:
                continue
            config_values[k] = v

        add_credential_process = os.environ.get('AWS_CONFIGURE_SSO_DISABLE_CREDENTIAL_PROCESS', '').lower() not in ['1', 'true']
        if credential_process is not None:
            add_credential_process = credential_process

        if add_credential_process:
            config_values["credential_process"] = "aws-sso-util credential-process --profile {}".format(config.profile_name)
        else:
            config_values.pop("credential_process", None)

        LOGGER.debug("Config values for profile {}: {}".format(config.profile_name, config_values))

        write_config(config.profile_name, config_values)


if __name__ == "__main__":
    populate_profiles(prog_name="python -m aws_sso_util.cli.populate_profiles")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import subprocess
import sys

import click

@click.command('configure-profile')
@click.argument('profile')
@click.option('--sso-start-url')
@click.option('--sso-region')
@click.option("--credential-process/--no-credential-process", default=None)
def configure_profile(
        profile,
        sso_start_url,
        sso_region,
        credential_process):
    def get(name):
        return subprocess.run(['aws', 'configure', 'get', 'profile.{}.{}'.format(profile, name)], stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout

    def set(name, value):
        subprocess.run(['aws', 'configure', 'set', 'profile.{}.{}'.format(profile, name), value or ''], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)

    default_sso_start_url =  os.environ.get('AWS_CONFIGURE_SSO_DEFAULT_SSO_START_URL',  os.environ.get('AWS_CONFIGURE_DEFAULT_SSO_START_URL'))
    default_sso_region    =  os.environ.get('AWS_CONFIGURE_SSO_DEFAULT_SSO_REGION',     os.environ.get('AWS_CONFIGURE_DEFAULT_SSO_REGION'))

    if sso_start_url:
        set('sso_start_url', sso_start_url)
    elif not get('sso_start_url') and default_sso_start_url:
        set('sso_start_url', default_sso_start_url)

    if sso_region:
        set('sso_region', sso_region)
    elif not get('sso_region') and default_sso_region:
        set('sso_region', default_sso_region)

    result = subprocess.run('aws configure sso --profile {}'.format(profile), shell=True).returncode

    if result:
        sys.exit(result)

    add_credential_process = os.environ.get('AWS_CONFIGURE_SSO_DISABLE_CREDENTIAL_PROCESS', '').lower() not in ['1', 'true']
    if credential_process is not None:
        add_credential_process = credential_process

    if add_credential_process:
        credential_process_opts = ''
        set('credential_process', 'aws-sso-util credential-process --profile {profile}{opts}'.format(profile=profile, opts=credential_process_opts))

if __name__ == '__main__':
    configure_profile(prog_name="python -m aws_sso_util.cli.configure_profile")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter

